import salt.utils.platform
import salt.utils.url
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class UrlTestCase(TestCase):
    """
    TestCase for salt.utils.url module
    """

    # parse tests

    def test_parse_path(self):
        """
        Test parsing an ordinary path
        """
        path = "interesting?/path&.conf:and other things"

        self.assertEqual(salt.utils.url.parse(path), (path, None))

    def test_parse_salt_url(self):
        """
        Test parsing a 'salt://' URL
        """
        path = "?funny/path with {interesting|chars}"
        url = "salt://" + path
        if salt.utils.platform.is_windows():
            path = "_funny/path with {interesting_chars}"

        self.assertEqual(salt.utils.url.parse(url), (path, None))

    def test_parse_salt_saltenv(self):
        """
        Test parsing a 'salt://' URL with a '?saltenv=' query
        """
        saltenv = "ambience"
        path = "?funny/path&with {interesting|chars}"
        url = "salt://" + path + "?saltenv=" + saltenv
        if salt.utils.platform.is_windows():
            path = "_funny/path&with {interesting_chars}"

        self.assertEqual(salt.utils.url.parse(url), (path, saltenv))

    # create tests

    def test_create_url(self):
        """
        Test creating a 'salt://' URL
        """
        path = "? interesting/&path.filetype"
        url = "salt://" + path
        if salt.utils.platform.is_windows():
            url = "salt://_ interesting/&path.filetype"

        self.assertEqual(salt.utils.url.create(path), url)

    def test_create_url_saltenv(self):
        """
        Test creating a 'salt://' URL with a saltenv
        """
        saltenv = "raumklang"
        path = "? interesting/&path.filetype"
        if salt.utils.platform.is_windows():
            path = "_ interesting/&path.filetype"

        url = "salt://" + path + "?saltenv=" + saltenv

        self.assertEqual(salt.utils.url.create(path, saltenv), url)

    def test_create_url_with_backslash_in_path(self):
        """
        Test creating a 'salt://' URL
        """
        src_path = r"? interesting\&path.filetype"
        tgt_path = "? interesting/&path.filetype"
        url = "salt://" + tgt_path
        if salt.utils.platform.is_windows():
            url = "salt://_ interesting/&path.filetype"

        self.assertEqual(salt.utils.url.create(src_path), url)

    def test_create_url_saltenv_with_backslash_in_path(self):
        """
        Test creating a 'salt://' URL with a saltenv
        """
        saltenv = "raumklang"
        src_path = r"? interesting\&path.filetype"
        tgt_path = "? interesting/&path.filetype"
        if salt.utils.platform.is_windows():
            tgt_path = "_ interesting/&path.filetype"

        url = "salt://" + tgt_path + "?saltenv=" + saltenv

        self.assertEqual(salt.utils.url.create(src_path, saltenv), url)

    # is_escaped tests

    def test_is_escaped_windows(self):
        """
        Test not testing a 'salt://' URL on windows
        """
        url = "salt://dir/file.ini"

        with patch("salt.utils.platform.is_windows", MagicMock(return_value=True)):
            self.assertFalse(salt.utils.url.is_escaped(url))

    def test_is_escaped_escaped_path(self):
        """
        Test testing an escaped path
        """
        path = "|dir/file.conf?saltenv=basic"

        self.assertTrue(salt.utils.url.is_escaped(path))

    def test_is_escaped_unescaped_path(self):
        """
        Test testing an unescaped path
        """
        path = "dir/file.conf"

        self.assertFalse(salt.utils.url.is_escaped(path))

    def test_is_escaped_escaped_url(self):
        """
        Test testing an escaped 'salt://' URL
        """
        url = "salt://|dir/file.conf?saltenv=basic"

        self.assertTrue(salt.utils.url.is_escaped(url))

    def test_is_escaped_unescaped_url(self):
        """
        Test testing an unescaped 'salt://' URL
        """
        url = "salt://dir/file.conf"

        self.assertFalse(salt.utils.url.is_escaped(url))

    def test_is_escaped_generic_url(self):
        """
        Test testing an unescaped 'salt://' URL
        """
        url = "https://gentoo.org/"

        self.assertFalse(salt.utils.url.is_escaped(url))

    # escape tests

    def test_escape_windows(self):
        """
        Test not escaping a 'salt://' URL on windows
        """
        url = "salt://dir/file.ini"

        with patch("salt.utils.platform.is_windows", MagicMock(return_value=True)):
            self.assertEqual(salt.utils.url.escape(url), url)

    def test_escape_escaped_path(self):
        """
        Test escaping an escaped path
        """
        resource = "|dir/file.conf?saltenv=basic"

        self.assertEqual(salt.utils.url.escape(resource), resource)

    def test_escape_unescaped_path(self):
        """
        Test escaping an unescaped path
        """
        path = "dir/file.conf"
        escaped_path = "|" + path
        if salt.utils.platform.is_windows():
            escaped_path = path

        self.assertEqual(salt.utils.url.escape(path), escaped_path)

    def test_escape_escaped_url(self):
        """
        Test testing an escaped 'salt://' URL
        """
        url = "salt://|dir/file.conf?saltenv=basic"

        self.assertEqual(salt.utils.url.escape(url), url)

    def test_escape_unescaped_url(self):
        """
        Test testing an unescaped 'salt://' URL
        """
        path = "dir/file.conf"
        url = "salt://" + path
        escaped_url = "salt://|" + path
        if salt.utils.platform.is_windows():
            escaped_url = url

        self.assertEqual(salt.utils.url.escape(url), escaped_url)

    def test_escape_generic_url(self):
        """
        Test testing an unescaped 'salt://' URL
        """
        url = "https://gentoo.org/"

        self.assertEqual(salt.utils.url.escape(url), url)

    # unescape tests

    def test_unescape_windows(self):
        """
        Test not escaping a 'salt://' URL on windows
        """
        url = "salt://dir/file.ini"

        with patch("salt.utils.platform.is_windows", MagicMock(return_value=True)):
            self.assertEqual(salt.utils.url.unescape(url), url)

    def test_unescape_escaped_path(self):
        """
        Test escaping an escaped path
        """
        resource = "dir/file.conf?saltenv=basic"
        escaped_path = "|" + resource

        self.assertEqual(salt.utils.url.unescape(escaped_path), resource)

    def test_unescape_unescaped_path(self):
        """
        Test escaping an unescaped path
        """
        path = "dir/file.conf"

        self.assertEqual(salt.utils.url.unescape(path), path)

    def test_unescape_escaped_url(self):
        """
        Test testing an escaped 'salt://' URL
        """
        resource = "dir/file.conf?saltenv=basic"
        url = "salt://" + resource
        escaped_url = "salt://|" + resource

        self.assertEqual(salt.utils.url.unescape(escaped_url), url)

    def test_unescape_unescaped_url(self):
        """
        Test testing an unescaped 'salt://' URL
        """
        url = "salt://dir/file.conf"

        self.assertEqual(salt.utils.url.unescape(url), url)

    def test_unescape_generic_url(self):
        """
        Test testing an unescaped 'salt://' URL
        """
        url = "https://gentoo.org/"

        self.assertEqual(salt.utils.url.unescape(url), url)

    # add_env tests

    def test_add_env_not_salt(self):
        """
        Test not adding a saltenv to a non 'salt://' URL
        """
        saltenv = "higgs"
        url = "https://pdg.lbl.gov/"

        self.assertEqual(salt.utils.url.add_env(url, saltenv), url)

    def test_add_env(self):
        """
        Test adding a saltenv to a 'salt://' URL
        """
        saltenv = "erstwhile"
        url = "salt://salted/file.conf"
        url_env = url + "?saltenv=" + saltenv

        self.assertEqual(salt.utils.url.add_env(url, saltenv), url_env)

    # split_env tests

    def test_split_env_non_salt(self):
        """
        Test not splitting a saltenv from a non 'salt://' URL
        """
        saltenv = "gravitodynamics"
        url = "https://arxiv.org/find/all/?" + saltenv

        self.assertEqual(salt.utils.url.split_env(url), (url, None))

    def test_split_env(self):
        """
        Test splitting a 'salt://' URL
        """
        saltenv = "elsewhere"
        url = "salt://salted/file.conf"
        url_env = url + "?saltenv=" + saltenv

        self.assertEqual(salt.utils.url.split_env(url_env), (url, saltenv))

    # validate tests

    def test_validate_valid(self):
        """
        Test URL valid validation
        """
        url = "salt://config/file.name?saltenv=vapid"
        protos = ["salt", "pepper", "cinnamon", "melange"]

        self.assertTrue(salt.utils.url.validate(url, protos))

    def test_validate_invalid(self):
        """
        Test URL invalid validation
        """
        url = "cumin://config/file.name?saltenv=vapid"
        protos = ["salt", "pepper", "cinnamon", "melange"]

        self.assertFalse(salt.utils.url.validate(url, protos))

    # strip tests

    def test_strip_url_with_scheme(self):
        """
        Test stripping of URL scheme
        """
        scheme = "git+salt+rsync+AYB://"
        resource = "all/the/things.stuff;parameter?query=I guess"
        url = scheme + resource

        self.assertEqual(salt.utils.url.strip_proto(url), resource)

    def test_strip_url_without_scheme(self):
        """
        Test stripping of a URL without a scheme
        """
        resource = "all/the/things.stuff;parameter?query=I guess"

        self.assertEqual(salt.utils.url.strip_proto(resource), resource)

    def test_http_basic_auth(self):
        """
        Tests that adding basic auth to a URL works as expected
        """
        # ((user, password), expected) tuples
        test_inputs = (
            ((None, None), "http://example.com"),
            (("user", None), "http://user@example.com"),
            (("user", "pass"), "http://user:pass@example.com"),
        )
        for (user, password), expected in test_inputs:
            kwargs = {
                "url": "http://example.com",
                "user": user,
                "password": password,
            }
            # Test http
            result = salt.utils.url.add_http_basic_auth(**kwargs)
            self.assertEqual(result, expected)
            # Test https
            kwargs["url"] = kwargs["url"].replace("http://", "https://", 1)
            expected = expected.replace("http://", "https://", 1)
            result = salt.utils.url.add_http_basic_auth(**kwargs)
            self.assertEqual(result, expected)

    def test_http_basic_auth_https_only(self):
        """
        Tests that passing a non-https URL with https_only=True will raise a
        ValueError.
        """
        kwargs = {
            "url": "http://example.com",
            "user": "foo",
            "password": "bar",
            "https_only": True,
        }
        self.assertRaises(ValueError, salt.utils.url.add_http_basic_auth, **kwargs)

    def test_redact_http_basic_auth(self):
        sensitive_outputs = (
            "https://deadbeaf@example.com",
            "https://user:pw@example.com",
        )
        sanitized = "https://<redacted>@example.com"
        for sensitive_output in sensitive_outputs:
            result = salt.utils.url.redact_http_basic_auth(sensitive_output)
            self.assertEqual(result, sanitized)

    def test_redact_non_auth_output(self):
        non_auth_output = "This is just normal output"
        self.assertEqual(
            non_auth_output, salt.utils.url.redact_http_basic_auth(non_auth_output)
        )



"""
Work with Nix packages
======================

.. versionadded:: 2017.7.0

Does not require the machine to be Nixos, just have Nix installed and available
to use for the user running this command. Their profile must be located in
their home, under ``$HOME/.nix-profile/``, and the nix store, unless specially
set up, should be in ``/nix``. To easily use this with multiple users or a root
user, set up the `nix-daemon`_.

This module exposes most of the common nix operations. Currently not meant to be run as a ``pkg`` module, but explicitly as ``nix.*``.

For more information on nix, see the `nix documentation`_.

.. _`nix documentation`: https://nixos.org/nix/manual/
.. _`nix-daemon`: https://nixos.org/nix/manual/#ssec-multi-user
"""


import itertools
import logging
import os

import salt.utils.itertools
import salt.utils.path

logger = logging.getLogger(__name__)


def __virtual__():
    """
    This only works if we have access to nix-env
    """
    nixhome = os.path.join(
        os.path.expanduser("~{}".format(__opts__["user"])), ".nix-profile/bin/"
    )
    if salt.utils.path.which(
        os.path.join(nixhome, "nix-env")
    ) and salt.utils.path.which(os.path.join(nixhome, "nix-collect-garbage")):
        return True
    else:
        return (
            False,
            "The `nix` binaries required cannot be found or are not installed."
            " (`nix-store` and `nix-env`)",
        )


def _run(cmd):
    """
    Just a convenience function for ``__salt__['cmd.run_all'](cmd)``
    """
    return __salt__["cmd.run_all"](
        cmd, env={"HOME": os.path.expanduser("~{}".format(__opts__["user"]))}
    )


def _nix_env():
    """
    nix-env with quiet option. By default, nix is extremely verbose and prints the build log of every package to stderr. This tells nix to
    only show changes.
    """
    nixhome = os.path.join(
        os.path.expanduser("~{}".format(__opts__["user"])), ".nix-profile/bin/"
    )
    return [os.path.join(nixhome, "nix-env")]


def _nix_collect_garbage():
    """
    Make sure we get the right nix-store, too.
    """
    nixhome = os.path.join(
        os.path.expanduser("~{}".format(__opts__["user"])), ".nix-profile/bin/"
    )
    return [os.path.join(nixhome, "nix-collect-garbage")]


def _quietnix():
    """
    nix-env with quiet option. By default, nix is extremely verbose and prints the build log of every package to stderr. This tells nix to
    only show changes.
    """
    p = _nix_env()
    p.append("--no-build-output")
    return p


def _zip_flatten(x, ys):
    """
    intersperse x into ys, with an extra element at the beginning.
    """
    return itertools.chain.from_iterable(zip(itertools.repeat(x), ys))


def _output_format(out, operation):
    """
    gets a list of all the packages that were affected by ``operation``, splits it up (there can be multiple packages on a line), and then
    flattens that list. We make it to a list for easier parsing.
    """
    return [s.split()[1:] for s in out if s.startswith(operation)]


def _format_upgrade(s):
    """
    split the ``upgrade`` responses on ``' to '``
    """
    return s.split(" to ")


def _strip_quotes(s):
    """
    nix likes to quote itself in a backtick and a single quote. This just strips those.
    """
    return s.strip("'`")


def upgrade(*pkgs):
    """
    Runs an update operation on the specified packages, or all packages if none is specified.

    :type pkgs: list(str)
    :param pkgs:
        List of packages to update

    :return: The upgraded packages. Example element: ``['libxslt-1.1.0', 'libxslt-1.1.10']``
    :rtype: list(tuple(str, str))

    .. code-block:: bash

        salt '*' nix.update
        salt '*' nix.update pkgs=one,two
    """
    cmd = _quietnix()
    cmd.append("--upgrade")
    cmd.extend(pkgs)

    out = _run(cmd)

    upgrades = [
        _format_upgrade(s.split(maxsplit=1)[1])
        for s in out["stderr"].splitlines()
        if s.startswith("upgrading")
    ]

    return [[_strip_quotes(s_) for s_ in s] for s in upgrades]


def install(*pkgs, **kwargs):
    """
    Installs a single or multiple packages via nix

    :type pkgs: list(str)
    :param pkgs:
        packages to update
    :param bool attributes:
        Pass the list of packages or single package as attribues, not package names.
        default: False

    :return: Installed packages. Example element: ``gcc-3.3.2``
    :rtype: list(str)

    .. code-block:: bash

        salt '*' nix.install package [package2 ...]
        salt '*' nix.install attributes=True attr.name [attr.name2 ...]
    """

    attributes = kwargs.get("attributes", False)

    if not pkgs:
        return "Plese specify a package or packages to upgrade"

    cmd = _quietnix()
    cmd.append("--install")

    if kwargs.get("attributes", False):
        cmd.extend(_zip_flatten("--attr", pkgs))
    else:
        cmd.extend(pkgs)

    out = _run(cmd)

    installs = list(
        itertools.chain.from_iterable(
            [
                s.split()[1:]
                for s in out["stderr"].splitlines()
                if s.startswith("installing")
            ]
        )
    )

    return [_strip_quotes(s) for s in installs]


def list_pkgs(installed=True, attributes=True):
    """
    Lists installed packages. Due to how nix works, it defaults to just doing a ``nix-env -q``.

    :param bool installed:
        list only installed packages. This can be a very long list (12,000+ elements), so caution is advised.
        Default: True

    :param bool attributes:
        show the attributes of the packages when listing all packages.
        Default: True

    :return: Packages installed or available, along with their attributes.
    :rtype: list(list(str))

    .. code-block:: bash

        salt '*' nix.list_pkgs
        salt '*' nix.list_pkgs installed=False
    """

    # We don't use -Q here, as it obfuscates the attribute names on full package listings.
    cmd = _nix_env()
    cmd.append("--query")

    if installed:
        # explicitly add this option for consistency, it's normally the default
        cmd.append("--installed")
    if not installed:
        cmd.append("--available")
        # We only show attributes if we're not doing an `installed` run.
        # The output of `nix-env -qaP` and `nix-env -qP` are vastly different:
        #    `nix-env -qaP` returns a list such as 'attr.path  name-version'
        #    `nix-env -qP` returns a list of 'installOrder  name-version'
        # Install order is useful to unambiguously select packages on a single
        # machine, but on more than one it can be a bad thing to specify.
        if attributes:
            cmd.append("--attr-path")

    out = _run(cmd)

    return [s.split() for s in salt.utils.itertools.split(out["stdout"], "\n")]


def uninstall(*pkgs):
    """
    Erases a package from the current nix profile. Nix uninstalls work differently than other package managers, and the symlinks in the
    profile are removed, while the actual package remains. There is also a ``nix.purge`` function, to clear the package cache of unused
    packages.

    :type pkgs: list(str)
    :param pkgs:
        List, single package to uninstall

    :return: Packages that have been uninstalled
    :rtype: list(str)

    .. code-block:: bash

        salt '*' nix.uninstall pkg1 [pkg2 ...]
    """

    cmd = _quietnix()
    cmd.append("--uninstall")
    cmd.extend(pkgs)

    out = _run(cmd)

    fmtout = out["stderr"].splitlines(), "uninstalling"

    return [
        _strip_quotes(s.split()[1])
        for s in out["stderr"].splitlines()
        if s.startswith("uninstalling")
    ]


def collect_garbage():
    """
    Completely removed all currently 'uninstalled' packages in the nix store.

    Tells the user how many store paths were removed and how much space was freed.

    :return: How much space was freed and how many derivations were removed
    :rtype: str

    .. warning::
       This is a destructive action on the nix store.

    .. code-block:: bash

        salt '*' nix.collect_garbage
    """
    cmd = _nix_collect_garbage()
    cmd.append("--delete-old")

    out = _run(cmd)

    return out["stdout"].splitlines()
